////////////////////////////////////////////////////////////////////////
#define TEST_5		// <-- select model here
#define  SAVE_RESULTS      // <-- specify whether to save results to file for plotting 
unsigned long samples_r   = 30;    // <-- select number of samples keep one order of magnitude less than phi
unsigned long samples_phi = 400;   // <-- select number of samples
#undef USE_PROFIL	// <-- specify to use PROFIL for interval arithmetic
#undef USE_FILIB	// <-- specify to use FILIB++ for interval arithmetic
////////////////////////////////////////////////////////////////////////

#define _DEBUG
#define _TRACE
#include "ffunc.hpp"
#include <fstream>
#include <iomanip>
#include <sys/time.h>

#include "ellimage.hpp"

#ifdef USE_PROFIL
  #include "mcprofil.hpp"
  typedef INTERVAL I;
#else
  #ifdef USE_FILIB
    #include "mcfilib.hpp"
    typedef filib::interval<double> I;
  #else
    #include "interval.hpp"
    typedef mc::Interval I;
  #endif
#endif

#include "ffunc.hpp"
#include <fstream>
#include <iomanip>
#include <algorithm>
#include <sys/time.h>

#include "ellimage.hpp"
#include "interval.hpp"
typedef mc::Interval    I     ;
typedef mc::EllImg<I>   EI    ;
typedef mc::EllVar<I>   EV    ;
typedef mc::Ellipsoid   E     ;
typedef CPPL::dcovector dcv   ;
typedef CPPL::dsymatrix dsm   ;


#if defined( TEST_1 ) 
const int NX = 2;	
const int NF = 2;	
dcv q0(NX);
dsm Q0(NX);
template <class T>
T f
( const T*x, const long i )
{
  switch( i ){
    case 0:  return x[0]*x[1];
    case 1:  return x[1]*x[0];
    default: throw std::runtime_error("invalid index");
  }
}

#elif defined( TEST_2)
const int NX = 2;	
const int NF = 2;	
dcv q0(NX);
dsm Q0(NX);
template <class T>
T f
( const T*x, const long i )
{
  using mc::sqr;
  switch( i ){
    case 0:  return sqr(x[0]*x[1]) + log(x[1]);
    case 1:  return log(x[0]);
    default: throw std::runtime_error("invalid index");
  }
}

#elif defined( TEST_3)
const int NX = 2;	
const int NF = 2;	
dcv q0(NX);
dsm Q0(NX);
template <class T>
T f
( const T*x, const long i )
{
  using mc::sqr;
  switch( i ){
    case 0:  return log(x[0]) + sqr(x[1]);
    case 1:  return sin(x[0]) - cos(x[1]);
    default: throw std::runtime_error("invalid index");
  }
}

#elif defined( TEST_4 )
const int NX = 2;	
const int NF = 2;	
dcv q0(NX);
dsm Q0(NX);
template <class T>
T f
( const T*x, const long i )
{
  using mc::sqr;
  switch( i ){
    case 0:  return  sqr(x[0]*x[1]) + log(x[1]);
    case 1:  return  sin(x[0]) * cos(x[1]);
    default: throw std::runtime_error("invalid index");
  }
}

#elif defined( TEST_5 )
const int NX = 2;	
const int NF = 2;	
dcv q0(NX);
dsm Q0(NX);
template <class T>
T f
( const T*x, const long i )
{
  using mc::sqr;
  switch( i ){
    case 0:  return  x[0] + x[1];
    case 1:  return  x[0] - x[1];
    default: throw std::runtime_error("invalid index");
  }
}

#endif

int main()
{
  mc::FFGraph FF;
  // Independent variables
  mc::FFVar X[NX];
  std::vector<const mc::FFVar*> vX;
  for( unsigned int i=0; i<NX; ++i) 
  {
  	X[i].set( &FF, i );
  	vX.push_back( &X[i] );
  }
  
  // dependent variables 
  mc::FFVar F[NF]; 
  std::vector<const mc::FFVar*> vF;
  for( unsigned int i=0; i<NF; ++i) 
  {
  	F[i] = f( X, i );
  	vF.push_back( &F[i] );
  } 
  CPPL::dssmatrix depmap = FF.generate_depmap( vF, vX );  
  std::cout << "\n DEPENDENCY MAP : \n" << depmap << std::endl;
 
  #if defined( TEST_1 ) || defined( TEST_2 ) || defined( TEST_3 ) || defined( TEST_4 ) || defined( TEST_5 )
  //center         shape
  q0(0) = 3.;     Q0(0,0) = 5.;
  q0(1) = 4.;     Q0(1,0) = 4.; Q0(1,1) = 5.; 
  #endif
  
  E Ex( Q0, q0 );	// Constructor for Ellipsoidal domain
  EV EX[NX];		// Constructor for Ellipsoidal independent variables
  EV EF[NF];		// Constructor for Ellipsoidal dependent variables
  EI Eflift;	// Constructor for lifted Ellipsoidal Image

  // Options
  Eflift.options.PREALLOCATE     = 0;
  Eflift.options.CHEBYSHEV_USE   = false;

  // Set Ellipsoidal domain in lifted ellipsoid 
  Eflift.set( Q0, q0 , depmap ); 
  // Set independent variables in the model
  for( long i=0 ; i<NF ; ++i ) EX[i].set( &Eflift , i ); 
  // Evaluate dependent variables
  for( long i=0 ; i<NF ; ++i ) EF[i] = f( EX, i );

  // Print lifted ellipsoid
  std::cout << "Lifted ellipsoidal image:\n" << Eflift << std::endl;
 
  mc::Ellipsoid Ef;

  if( NF == 1 ){
    // Project lifted ellipsoidal image
    EV XF[NX+NF] = { EX[0], EX[1], EF[0] };
    Ef = Eflift.get( XF, NX+NF ) ;
  } 
  else
  Ef = Eflift.get( EF, NF ) ;

  std::cout <<"Ellipsoidal Image:\n" << Ef << std::endl;


  #ifdef SAVE_RESULTS
  // Parameters for sampling
  I      radius    = I(0.,1.);
  I      phi       = I(0.,2.*mc::PI);
  double discr     = (mc::Op<I>::u(radius) - mc::Op<I>::l(radius) )/(double ) samples_r;
  double discphi   = (mc::Op<I>::u(phi)    - mc::Op<I>::l(phi)    )/(double ) samples_phi;
  double phiEr_sam = mc::Op<I>::l(phi);

  std::ostringstream name1;
  name1 << "Fsam.out" ; 
  std::ofstream fout(name1.str().c_str() , std::ios_base::out );
  fout << std::scientific << std::setprecision(8) << std::right;

  std::ostringstream name2;
  name2 << "Esam.out"; 
  std::ofstream eout(name2.str().c_str() , std::ios_base::out );
  eout << std::scientific << std::setprecision(8) << std::right;    

  CPPL::dcovector xy_vec(NF), z_sam(NF), E_sam(NF), Fimg(NF) ;
  while(phiEr_sam <= mc::Op<I>::u(phi)){
    xy_vec(0) = mc::Op<I>::u(radius) * cos(phiEr_sam);
    xy_vec(1) = mc::Op<I>::u(radius) * sin(phiEr_sam);
    E_sam     = Ef.c() + Ef.sqrtQ()*xy_vec;
    //print
    eout  <<   E_sam(0) << '\t' <<  E_sam(1) << std::endl;

    double  rad_sam = mc::Op<I>::u(radius);
    while( rad_sam >= mc::Op<I>::l(radius) +discr ){	
      double  phi_sam = mc::Op<I>::l(phi);
      while(phi_sam <= mc::Op<I>::u(phi) ){
        xy_vec(0) = rad_sam * cos(phi_sam);
        xy_vec(1) = rad_sam * sin(phi_sam);
        z_sam     = Ex.c() + Ex.sqrtQ()*xy_vec;
        // Evaluate dependent variables
        double z_d[2] ={z_sam(0),z_sam(1)};
        for( unsigned long i=0; i<NX; ++i ) Fimg(i) = f( z_d, i );
        //print
        if( phiEr_sam == mc::Op<I>::l(phi) ) fout <<  Fimg(0) << '\t' << Fimg(1) << std::endl;
        phi_sam += discphi;
      }
      rad_sam -= discr;
    }
    // Evaluate at center
    double z_c[2] ={Ex.c()(0), Ex.c()(1)};
    for( unsigned long i=0; i<NX; ++i ) Fimg(i) = f( z_c , i );
    //print evaluation at center
    if( phiEr_sam == mc::Op<I>::l(phi) ) fout <<  Fimg(0) << '\t' << Fimg(1) << std::endl;

    phiEr_sam += discphi;
  }
  #endif
  
  
  return 0;
}





