#define TEST_1	// <-- select test function here
const int NTE = 4;	// <-- select Taylor expansion order here
const int NX = 500;	// <-- select X discretization here
#define SAVE_RESULTS    // <-- specify whether to save results to file
#undef USE_PROFIL	// <-- specify to use PROFIL for interval arithmetic
#undef USE_FILIB	// <-- specify to use FILIB++ for interval arithmetic

////////////////////////////////////////////////////////////////////////

#include <fstream>
#include <iomanip>

#include "interval.hpp"
#include "tmodel.hpp"
typedef mc::Interval I;
typedef mc::TModel<I> PM;
typedef mc::TVar<I> PV;

using namespace std;
using namespace mc;

////////////////////////////////////////////////////////////////////////

#if defined( TEST_1 )
const double XL   = -1.;	// <-- X range lower bound
const double XU   =  2.;	// <-- X range upper bound
template <class T>
T myfunc
( const T&x )
{
  return 1. - 5.*x - pow(x,2)/2. + pow(x,3)/3. + pow(x,5)/9.;
}
#endif

////////////////////////////////////////////////////////////////////////
int main()
////////////////////////////////////////////////////////////////////////
{


  ofstream res( "TEST.out", ios_base::out );
  res << std::scientific << std::setprecision(5) << std::right;


 // Define Chebyshev model environment
    PM mod( 1, NTE );

    // Define variable X, and evaluate Chebyshev model
    PV CVX( &mod, 0, I(XL,XU) );
    PV CVF = myfunc( CVX );
    std::cout << "\nPolynomial model of f(x):" << CVF;

    // Repeated calculations at grid points (for display)
    for( int iX=0; iX<NX; iX++ ){ 

      double DX = XL+iX*(XU-XL)/(NX-1.);
      double DF = myfunc( DX );
      CVX.set( &mod, 0, I(XL,XU) );
      CVF = myfunc( CVX );
      I BF = CVF.B();
      double PF = CVF.P( &DX );
      I IF = PF + CVF.R();


#ifdef SAVE_RESULTS
      res << std::setw(14) << DX << std::setw(14) << std::setw(14) << DF
          << std::setw(14) << PF << std::setw(14) << IF.l()  << setw(14) << IF.u()
          << std::setw(14) << BF.l()  << std::setw(14) << BF.u()
          << std::endl;
#endif
    }

return 0;

    }

  

